#include "common.h"
#include "dlist.h"

/* null structure to easy initializing to NULL */
const DList DListNULL = { NULL, NULL };

/* create a new dlist node containing data */
DListEntry* 
dlist_new(void* data) {
  DListEntry* new = XMALLOC( DListEntry, 1 );
  
  new->next = NULL;
  new->prev = NULL;
  new->data = data;

  return new;
}

/* concatenate two lists */
void
dlist_cat(DList* list1, DList* list2){
  list1->tail->next = list2->head;
  list2->head->prev = list1->tail;
  list1->tail = list2->tail;
}

/* to shorten things a bit in some cases */
DListEntry*
dlist_insertnew(DList* list, void* data, DListEntry* pos) {
  DListEntry* new = dlist_new( data );
  return dlist_insert( list, new, pos );
}

/* can be used to insert an entry at a specific position in a dlist.
   use with caution on empty list, cause it might fail then. */
DListEntry*
dlist_insert(DList* list, DListEntry* new, DListEntry* pos) {

  if ( pos ) { // if pos not null, insert
    new->next = pos->next;
    pos->next = new;
    new->prev = pos;
  } else if( ! list->head && ! list->tail ) { // otherwise treat as empty
    new->next = NULL;                         // list
    new->prev = NULL;
    list->head = new;
  } else
    return NULL;    // not an empty list -> error
  
  if ( new->next )  // set tail when tail
    new->next->prev = new;
  else
    list->tail = new;
  
  return new;
}

/* can only be used to append to the end of a list */
DListEntry*
dlist_append(DList* list, DListEntry* new) {
  if ( list->tail ) { // if list tail exists, then link with it
    new->next = list->tail->next;
    list->tail->next = new;
  } else {
    new->next = NULL;
    list->head = new;
  }

  new->prev = list->tail;

  if ( new->next )  // if next is defined, then set its prev value correctly
    new->next->prev = new;
  else
    list->tail = new;
  
  return new;
}



DListEntry* 
dlist_insert_sort (DList* list, DListEntry* new, dlist_comp cmp) {
  
}

/* retrieves the head of a list */
DListEntry* 
dlist_head(DList* list) {
  DListEntry *help = list->tail;
  DListEntry *old  = NULL;
  
  while( help ) {
    old = help;
    help = help->prev;
  }
  
  return old;
}

/* retrieves the tail of a list */
DListEntry* 
dlist_tail(DList* list) {
  DListEntry *help = list->head;
  DListEntry *old  = NULL;

  while( help ) {
    old = help;
    help = help->next;
  }
  
  return old;
}

/* returns the length of the list */
size_t 
dlist_length(DList* list){
  DListEntry *help = list->head;
  size_t n;

  for (n = 0; help; ++n)
    help = help->next;

  return n;
}

/* removes an entry from the list, and returns it so that it might
   be destroyed properly */
DListEntry*
dlist_remove(DList* list, DListEntry* entry) {
  if (entry) {
    if (entry != list->head) {  // when not head
      if (entry->prev)       // and a previous entry exists
	entry->prev->next = entry->next; // tell the previous our next entry
    } else
      list->head = entry->next; // otherwise next entry is head
    
    if (entry != list->tail) {  // when not tail
      if (entry->next)       // and next entry exists
	entry->next->prev = entry->prev; // tell the next entry our previous 
    } else
    list->tail = entry->prev; // otherwise previous is tail
  }

  return entry;
}

/* checks a list to be ok (debugging) verifies head, tail and prints list */
dlist_check( DList* list ) {
  // should be implemented
}


DListEntry*
dlist_search(DList* list, void* data, dlist_comp cmp) 
{
  DListEntry *help = list->head;

  if (list) {
    while ( help ) {
      if ( cmp( data, help->data ) == 0 ) 
	return help;
      else
	help = help->next;
    }
  }
  
  return NULL;
}


int
dlist_strcmp(void* data1, void* data2)
{
  return strcmp( (char*) data1, (char*) data2 );
}

