#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include "common.h"
#include "error.h"

#include <libgen.h>  /* basename(3) */
const char *program_name = NULL;
      
static void error PARAMS((int exit_status, const char *mode,
			  const char *message, va_list ap));
      
void
set_program_name (path)
     const char *path;
{
  METAMAIL_ASSERT (!program_name);
  program_name = (char*) xstrdup ( (char*) basename (path));
}

static void
error (exit_status, mode, message, ap)
     int exit_status;
     const char *mode;
     const char *message;
     va_list ap;
{
  fprintf (stderr, "%s: %s: ", program_name, mode);
  vfprintf (stderr, message, ap);
  fprintf (stderr, ".\n");

  if (exit_status >= 0)
    exit (exit_status);
}

void
#if HAVE_STDARG_H
metamail_warning (const char *message, ...)
#else
metamail_warning (message, va_alist)
     const char *message;
     va_dcl
#endif
{
  va_list ap;
  VA_START (ap, message);
  error (-1, "warning", message, ap);
  va_end (ap);
}

void
#if HAVE_STDARG_H
metamail_error (const char *message, ...)
#else
metamail_error (message, va_alist)
     const char *message;
     va_dcl
#endif
{
  va_list ap;
  VA_START (ap, message);
  error (-1, "ERROR", message, ap);
  va_end (ap);
}
      
void
#if HAVE_STDARG_H
metamail_fatal (const char *message, ...)
#else
metamail_fatal (message, va_alist)
     const char *message;
     va_dcl
#endif
{
  va_list ap;
  VA_START (ap, message);
  error (EXIT_FAILURE, "FATAL", message, ap);
  va_end (ap);
}


