/*
Copyright (c) 1991 Bell Communications Research, Inc. (Bellcore)

Permission to use, copy, modify, and distribute this material 
for any purpose and without fee is hereby granted, provided 
that the above copyright notice and this permission notice 
appear in all copies, and that the name of Bellcore not be 
used in advertising or publicity pertaining to this 
material without the specific, prior written permission 
of an authorized representative of Bellcore.  BELLCORE 
MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY 
OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", 
WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES.
*/
/* mailcap.c - contains functions to access and parse mailcap files
    Authors:  Nathaniel S. Borenstein, Bellcore
              Michael Moerz <mikem@debian.org>
*/

#include <metamail/common.h>

#include "mailcap.h"
#include "metamail.h"

/* Only one or the other set up builtins gets used,
  depending on whether or not we're in the middle of 
          a multipart/alternative body part */
struct MailcapEntry BuiltIns[] = {
    {"text/*", CATTEMPLATE, NULL, 0, 1, 0, "plain text", LPRTEMPLATE},
    {NULL, NULL, NULL, 0, 0, 0}};

struct MailcapEntry BuiltInsAlternative[] = {
    {"text/plain", CATTEMPLATE, NULL, 0, 1, 0, "plain text", LPRTEMPLATE},
    {NULL, NULL, NULL, 0, 0, 0}};

/** hmm, ... 
*/
int
ProcessMailcapFiles(SquirrelFile) 
char *SquirrelFile;
{
    char *s, *pathcopy = NULL;
    char *path = getenv("MAILCAPS");
    if (!path) {
        uid_t uid = getuid();
        struct passwd *p;
        p = getpwuid(uid);
	if (p) {
	  path = XMALLOC(char, 5 + strlen(p->pw_dir) + strlen(STDPATH));
	}
        strcpy(path, p->pw_dir);
        strcat(path, STDPATH);
        pathcopy = path;
    } else 
    {
        pathcopy = XMALLOC(char, 1+strlen(path));
        strcpy(pathcopy, path);
        path = pathcopy;
    }
    while(path) {
        s = index(path, PATH_SEPARATOR);
        if (s) *s++ = 0;
        if (!ProcessMailcapFile(path, SquirrelFile)) return(0);
        path = s;
    }
    if (pathcopy) XFREE(pathcopy);
    return(-1);
}

int
TryBuiltIns(SquirrelFile) 
char *SquirrelFile;
{
  int i;
  /* Last resort -- for sites that didn't bother putting a "text" line in 
     their mailcap files... */
  if (DoDebug) 
    fprintf(stderr, "Looking for '%s' in built-in content-type handling "
	    "settings.\n", ContentType);
  for (i=0; BuiltIns[i].contenttype; ++i) {
    if (!TryMailcapEntry(SquirrelFile ? BuiltInsAlternative[i] : BuiltIns[i], SquirrelFile))
      return(0);
  }
  return(-1);
}

int
ProcessMailcapFile(file, SquirrelFile)
char *file, *SquirrelFile;
{
    struct MailcapEntry mc;
    FILE *fp;
    int res;

    res = IsDirectory(file);
    if (res) {
	if (res == 1) {
	    fprintf(stderr, "Mailcap path entry '%s' is a directory, should be a readable file.\n", file);
	} else if (DoDebug) {
	/* Quietly ignore non-existent mailcap files unless debugging */
	    fprintf(stderr, "Mailcap path entry '%s' does not exist.\n", file);
	}
	return(-1);
    }
    fp = fopen(file, "r");
    if (DoDebug) fprintf(stderr, "Looking for '%s' in mailcap file '%s'.\n", ContentType, file);
    while (fp && !feof(fp)) {
        mc.needtofree = 0;
	if (GetMailcapEntry(fp, &mc)) {
            if (!TryMailcapEntry(mc, SquirrelFile)) {
                fclose(fp);
                return(0);
            }
	}
    }
    if (fp) fclose(fp);
    return(-1);
}
